/*
    SPDX-FileCopyrightText: 2006 Pino Toscano <toscano.pino@tiscali.it>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef _DLGANNOTATIONS_H_
#define _DLGANNOTATIONS_H_

#include <QWidget>

class DlgAnnotations : public QWidget
{
    Q_OBJECT

public:
    explicit DlgAnnotations(QWidget *parent = nullptr);
};

#endif
