/*
    SPDX-FileCopyrightText: 2024 Pratham Gandhi <ppg.1382@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "js_global_p.h"

using namespace Okular;

JSGlobal::JSGlobal(QObject *parent)
    : QObject(parent)
{
}

JSGlobal::~JSGlobal() = default;