/*
    SPDX-FileCopyrightText: 2014 Laurent Montel <montel@kde.org>
    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef FAX_DEBUG_H
#define FAX_DEBUG_H

#include <QLoggingCategory>
Q_DECLARE_LOGGING_CATEGORY(FAX_LOG)

#endif
