<?xml version="1.0"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="okular_part" version="55">
<MenuBar>
  <Menu name="file"><text>&amp;File</text>
    <Action name="get_new_stuff" group="file_open"/>
    <Action name="import_ps" group="file_open"/>
    <Action name="file_save" group="file_save"/>
    <Action name="file_save_as" group="file_save"/>
    <Action name="file_reload" group="file_save"/>
    <Action name="file_print" group="file_print"/>
    <Action name="file_print_preview" group="file_print"/>
    <Action name="open_containing_folder" group="file_print"/>
    <Action name="properties" group="file_print"/>
    <Action name="embedded_files" group="file_print"/>
    <Action name="file_export_as" group="file_print"/>
    <Action name="file_share" group="file_print"/>
  </Menu>
  <Menu name="edit"><text>&amp;Edit</text>
    <Action name="edit_undo"/>
    <Action name="edit_redo"/>
    <Separator/>
    <Action name="edit_copy"/>
    <Action name="edit_copy_without_line_breaks"/>
    <Separator/>
    <Action name="edit_select_all"/>
    <Action name="edit_select_all_current_page"/>
    <Separator/>
    <Action name="edit_find"/>
    <Action name="edit_find_next"/>
    <Action name="edit_find_prev"/>
  </Menu>
  <Menu name="view"><text>&amp;View</text>
    <Action name="presentation"/>
    <Separator/>
    <Action name="view_zoom_out"/>
    <Action name="view_zoom_in"/>
    <Action name="view_actual_size"/>
    <Action name="view_fit_to_width"/>
    <Action name="view_fit_to_page"/>
    <Action name="view_auto_fit"/>
    <Separator/>
    <Action name="view_render_mode"/>
    <Separator/>
    <Menu name="view_orientation"><text>&amp;Orientation</text>
      <Action name="view_orientation_rotate_ccw"/>
      <Action name="view_orientation_rotate_cw"/>
      <Action name="view_orientation_original"/>
    </Menu>
    <Action name="view_trim_mode"/>
    <Separator/>
    <Action name="view_toggle_forms"/>
  </Menu>
  <Menu name="go"><text>&amp;Go</text>
    <Action name="go_previous"/>
    <Action name="go_next"/>
    <Separator/>
    <Action name="first_page"/>
    <Action name="last_page"/>
    <Separator/>
    <Action name="go_document_back"/>
    <Action name="go_document_forward" />
    <Separator/>
    <Action name="go_goto_page"/>
  </Menu>
  <Menu name="bookmarks"><text>&amp;Bookmarks</text>
    <Action name="bookmark_add"/>
    <Action name="rename_bookmark"/>
    <Action name="previous_bookmark"/>
    <Action name="next_bookmark" />
    <Separator/>
    <ActionList name="bookmarks_currentdocument" />
  </Menu>
  <Menu name="tools"><text>&amp;Tools</text>
    <Action name="mouse_drag"/>
    <Action name="mouse_zoom"/>
    <Action name="mouse_select"/>
    <Action name="mouse_textselect"/>
    <Action name="mouse_tableselect"/>
    <Action name="mouse_magnifier"/>
    <Separator/>
    <Action name="mouse_toggle_annotate"/>
    <Separator/>
    <Action name="add_digital_signature"/>
    <Separator/>
    <Action name="speak_document"/>
    <Action name="speak_from_current_page"/>
    <Action name="speak_current_page"/>
    <Action name="speak_stop_all"/>
    <Action name="speak_pause_resume"/>
  </Menu>
  <Menu name="settings"><text>&amp;Settings</text>
    <Action name="colorscheme_menu" group="show_merge"/>
    <Action name="show_leftpanel" group="show_merge"/>
    <Action name="show_bottombar" group="show_merge"/>
    <Action name="options_configure_generators" group="configure_merge"/>
    <Action name="options_configure" group="configure_merge"/>
  </Menu>
  <Menu name="help"><text>&amp;Help</text>
    <Action name="help_about_backend" group="about_merge"/>
  </Menu>
</MenuBar>
<ToolBar name="mainToolBar"><text>Main Toolbar</text>
  <Action name="show_leftpanel"/>
  <Separator/>
  <Action name="view_zoom_in"/>
  <Action name="view_zoom_out"/>
  <Action name="zoom_to" />
  <Action name="view_render_mode"/>
  <Spacer/>
  <Action name="page_number"/>
  <Spacer/>
  <Action name="mouse_drag"/>
  <Action name="mouse_selecttools"/>
  <Action name="annotation_favorites"/>
  <Separator/>
  <Action name="hamburger_menu"/>
</ToolBar>
<ActionProperties scheme="Default">
  <Action priority="0" name="show_leftpanel"/>
  <Action priority="0" name="view_zoom_out"/>
  <Action priority="0" name="view_zoom_in"/>
</ActionProperties>
<ToolBar name="annotationToolBar" iconText="icononly" newline="true" hidden="true"><text>Annotation Toolbar</text>
  <Action name="annotation_highlighter"/>
  <Action name="annotation_underline"/>
  <Action name="annotation_squiggle"/>
  <Action name="annotation_strike_out"/>
  <Action name="annotation_typewriter"/>
  <Action name="annotation_inline_note"/>
  <Action name="annotation_popup_note"/>
  <Action name="annotation_freehand_line"/>
  <Action name="annotation_geometrical_shape"/>
  <Action name="annotation_stamp"/>
  <Separator/>
  <Action name="annotation_settings_width"/>
  <Action name="annotation_settings_color"/>
  <Action name="annotation_settings_inner_color"/>
  <Action name="annotation_settings_opacity"/>
  <Action name="annotation_settings_font"/>
  <Action name="annotation_settings_advanced"/>
  <Separator/>
  <Action name="annotation_bookmark"/>
  <Action name="annotation_settings_pin"/>
  <Spacer/>
  <Action name="hide_annotation_toolbar"/>
</ToolBar>
<ToolBar name="quickAnnotationToolBar" iconText="icononly" newline="true" hidden="true" position="right" iconSize="32"><text>Quick Annotation Toolbar</text>
    <Action name="quick_annotation_action_bar"/>
</ToolBar>
</gui>
