
/***************************************************************************
 *   SPDX-FileCopyrightText: 2006 Luigi Toscano <luigi.toscano@tiscali.it>        *
 *   SPDX-FileCopyrightText: 2014 Frederik Gladhorn <gladhorn@kde.org>            *
 *                                                                         *
 *   SPDX-License-Identifier: GPL-2.0-or-later
 ***************************************************************************/

#ifndef OKULAR_DEBUG_MD_H
#define OKULAR_DEBUG_MD_H

#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(OkularMdDebug)

#endif
